class TextPiece {
	/**
	 * @param  {string}    text 
	 * @param  {...string} tags 
	 */
	constructor ( text, ...tags ) {
		this.text = text;
		this.tags = tags;
	}
}

class FormattingMatch {
	constructor () {
		this.excludeAny = false;
		this.excludeExact = false;
		this.includeAny = false;
		this.includeExact = false;
		this.regex = false;
		this.tagFunc = false;
		this.textFunc = false;
	}

	/**
	 * @param {TextPiece} piece
	 * @returns {boolean} 
	 */
	test ( piece ) {
		if ( typeof this.excludeAny !== 'boolean' ) {
			let copy = [ ...piece.tags ].filter( e => { return !this.excludeAny.includes( e ); } );
			if ( copy.length !== this.excludeAny.length ) return false;
		}
		if ( typeof this.excludeExact !== 'boolean' ) {
			if ( this.excludeExact.length === piece.tags.length ) {
				let copy = [ ...piece.tags ].filter( e => { return !this.excludeExact.includes( e ); } );
				if ( copy.length === 0 ) return false;
			}
		}
		if ( typeof this.includeAny !== 'boolean' ) {
			let copy = [ ...piece.tags ].filter( e => { return this.includeAny.includes( e ); } );
			if ( copy.length === 0 ) return false;
		}
		if ( typeof this.includeExact !== 'boolean' ) {
			if ( this.includeExact.length === piece.tags.length ) {
				let copy = [ ...piece.tags ].filter( e => { return !this.includeExact.includes( e ); } );
				if ( copy.length !== 0 ) return false;
			} else return false;
		}
		if ( typeof this.regex !== 'boolean' ) {
			if ( !this.regex.test( piece.text ) ) return false;
		}
		if ( typeof this.tagFunc !== 'boolean' ) {
			if ( !this.tagFunc( [ ...piece.tags ] ) ) return false;
		}
		if ( typeof this.textFunc !== 'boolean' ) {
			if ( !this.textFunc( piece.text ) ) return false;
		}
		return true;
	}

	/**
	 * @param  {...sting} tags 
	 */
	ExcludeAny ( ...tags ) {
		this.excludeAny = tags;
		return this;
	}
	/**
	 * @param  {...sting} tags 
	 */
	ExcludeExact ( ...tags ) {
		this.excludeExact = tags;
		return this;
	}
	/**
	 * @param  {...sting} tags 
	 */
	IncludeAny ( ...tags ) {
		this.includeAny = tags;
		return this;
	}
	/**
	 * @param  {...sting} tags 
	 */
	IncludeExact ( ...tags ) {
		this.includeExact = tags;
		return this;
	}
	regexTextMatch ( regex ) {
		this.regex = regex;
		return this;
	}
	/**
	 * @param {(text:string)=>boolean} callback 
	 */
	functionTextTest ( callback ) {
		this.textFunc = callback;
		return this;
	}
	/**
	 * @param {(tags:string[])=>boolean} callback 
	 */
	functionTagTest ( callback ) {
		this.tagFunc = callback;
		return this;
	}
}

class FormattingRule {
	constructor ( beginMatch ) {
		this.beginMatch = beginMatch;
		this.endMatch = false;
		this.nest = 0;
		this.cb = ( p ) => {};
	}
	EndMatch ( match ) {
		this.endMatch = match;
		return this;
	}
	/**
	 * whether this rule ends the last instance ( 0, default ), first instance ( 1 ), or all ( 2 )
	 * @param {number} type 
	 */
	Nest ( type ) {
		this.nest = type;
		return this;
	}
	/**
	 * @param {(text:TextPiece[])=>void} cb 
	 */
	callback ( cb ) {
		this.cb = cb;
		return this;
	}
}

class FormattingPass {
	constructor ( ...rules ) {
		this.rules = rules;
	}

	static selectNewLines ( textPieces ) {
		let all = [
			
		];
		let t = '';
		for ( let i = 0; i < textPieces.length; i++ ) {
			t += textPieces[ i ].text;
		}
		let te = t;
		t = t.split( /\r\n|\n\r|\n|\r/g );
		let ends = te.match( /\r\n|\n\r|\n|\r/g );
		let am = 0;
		for ( let i = 0; i < t.length; i++ ) {
			all.push( am );
			if ( ends.length !== 0 )
				t[ i ] += ends.shift();
			am += t[ i ].length;
		}
		am = 0;
		for ( let i = 0, n = 0; i < all.length; ) {
			if ( am + textPieces[ n ].text.length < all[ i ] ) {
				am += textPieces[ n++ ].text.length;
			} else {
				all[ i ] = { index: all[ i++ ] - am, piece: textPieces[ n ] };
			}
		}
		return all.reverse();
	}

	/**
	 * @param {TextPiece[]} textPieces
	 * @returns {TextPiece[]}
	 */
	run ( textPieces ) {
		let fwdinstances = [];
		let bwdinstances = [];
		for( let i = 0; i < textPieces.length; i++ ) {
			let piece = textPieces[ i ];
			this.rules.forEach( r => {
				if ( r.beginMatch.test( piece ) ) {
					if ( typeof r.endMatch === 'boolean' ) {
						r.cb( [ piece ] );
					} else {
						if ( r.nest === 0 ) {
							bwdinstances.push( { start: i, rule: r } );
						} else {
							fwdinstances.push( { start: i, rule: r } );
						}
					}
				}
			} );

			let exclude = [];
			for( let x = 0; x < fwdinstances.length; x++ ) {
				let r = fwdinstances[ x ];
				if ( exclude.includes( r.rule ) ) continue;
				if ( r.rule.endMatch.test( piece ) ) {
					if ( r.rule.nest !== 2 ) {
						exclude.push( r.rule );
					}
					r.rule.cb( textPieces.slice( r.start, i + 1 ) );
					fwdinstances.splice( fwdinstances.indexOf( r ), 1 );
				}
			}
			exclude = [];
			for( let x = bwdinstances.length - 1; x >= 0; x-- ) {
				let r = bwdinstances[ x ];
				if ( exclude.includes( r.rule ) ) continue;
				if ( r.rule.endMatch.test( piece ) ) {
					exclude.push( r.rule );
					r.rule.cb( textPieces.slice( r.start, i + 1 ) );
					bwdinstances.splice( bwdinstances.indexOf( r ), 1 );
				}
			}
		}
		return textPieces;
	}

	format ( textPieces ) {
		this.run( textPieces );
		let r = '';
		textPieces.forEach( t => {
			r += t.text;
		} );
		return r;
	}
}

class FormatCollection {
	constructor ( ...passes ) {
		this.passes = passes;
	}

	run ( textPieces ) {
		this.passes.forEach( p => {
			p.run( textPieces );
		} );

		return this.finish( textPieces );
	}

	finish ( textPieces ) {
		let r = '';
		textPieces.forEach( t => {
			r += t.text;
		} );
		return r;
	}
}

module.exports = {
	TextPiece: TextPiece,
	FormattingMatch: FormattingMatch,
	FormattingRule: FormattingRule,
	FormattingPass: FormattingPass,
	FormatCollection: FormatCollection
};